/* from include/linux/tracepoint.h */
#include <linux/tracepoint.h>

#define KLPR___DO_TRACE_CALL(name, args)   (*klpe___traceiter_##name)(NULL, args)

#define KLPR___DO_TRACE(name, args, cond, rcuidle)         \
	do {                                \
		int __maybe_unused __idx = 0;               \
                                   \
		if (!(cond))                        \
			return;                     \
		\
		/* srcu can't be used from NMI */           \
		WARN_ON_ONCE(rcuidle && in_nmi());          \
		\
		/* keep srcu and sched-rcu usage consistent */      \
		preempt_disable_notrace();              \
		\
		/*                          \
		 * For rcuidle callers, use srcu since sched-rcu    \
		 * doesn't work from the idle path.         \
		 */                         \
		if (rcuidle) {                      \
			__idx = srcu_read_lock_notrace(&tracepoint_srcu);\
			rcu_irq_enter_irqson();             \
		}                           \
		\
		KLPR___DO_TRACE_CALL(name, TP_ARGS(args));          \
		\
		if (rcuidle) {                      \
			rcu_irq_exit_irqson();              \
			srcu_read_unlock_notrace(&tracepoint_srcu, __idx);\
		}                           \
		\
		preempt_enable_notrace();               \
	} while (0)

#define KLPR___DECLARE_TRACE(name, proto, args, cond, data_proto)	\
	static int (*klpe___traceiter_##name)(data_proto);		\
	static struct tracepoint (*klpe___tracepoint_##name);		\
	static inline void klpr_trace_##name(proto)			\
	{								\
		if (static_key_enabled(&(*klpe___tracepoint_##name).key)) \
		KLPR___DO_TRACE(name,					\
				TP_ARGS(args),				\
				TP_CONDITION(cond), 0);			\
		if (IS_ENABLED(CONFIG_LOCKDEP) && (cond)) {		\
			rcu_read_lock_sched_notrace();			\
			rcu_dereference_sched((*klpe___tracepoint_##name).funcs); \
			rcu_read_unlock_sched_notrace();		\
		}							\
	}								\

#define KLPR_DECLARE_TRACE(name, proto, args)			\
	KLPR___DECLARE_TRACE(name, PARAMS(proto), PARAMS(args),	\
		cpu_online(raw_smp_processor_id()),		\
		PARAMS(void *__data, proto))

#define KLPR_DEFINE_EVENT(template, name, proto, args) \
	KLPR_DECLARE_TRACE(name, PARAMS(proto), PARAMS(args))

#define KLPR_TRACE_EVENT(name, proto, args) \
	KLPR_DECLARE_TRACE(name, PARAMS(proto), PARAMS(args))

/* klp-ccp: from fs/nfsd/nfs4proc.c */
#define KLPR_NFSDDBG_FACILITY  NFSDDBG_PROC

#undef klpr_ifdebug
# define klpr_ifdebug(fac) if (unlikely((*klpe_nfsd_debug) & KLPR_NFSDDBG_##fac))

# define klpr_dfprintk(fac, fmt, ...)			\
do {							\
	klpr_ifdebug(fac)				\
	printk(KERN_DEFAULT fmt, ##__VA_ARGS__);	\
} while (0)

#define klpr_dprintk(fmt, ...) \
	klpr_dfprintk(FACILITY, fmt, ##__VA_ARGS__)
